from time import time
from functools import wraps
from typing import Callable


def get_time(func: Callable) -> Callable:
    """
    装饰器，方便打印函数的执行时间。
    Args:
        func: 包装的函数
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time()
        ret = func(*args, **kwargs)
        use_time = time() - start_time

        arg_strs = [f'{arg}' for arg in args]
        arg_strs += [f'{k} = {v}' for k, v in kwargs.items()]
        print(f"\"{func.__name__}({', '.join(arg_strs)})\" took {use_time:.4f} seconds to execute.")
        return ret

    return wrapper



import base64
import hashlib
import hmac
import json
import time

from .base import BaseNotification
import requests


class DingtalkNotification(BaseNotification):
    """
    钉钉 Webhook 机器人。
    文档地址: https://open.dingtalk.com/document/orgapp/robot-message-types-and-data-format#title-z74-8to-i7e
    """

    def __init__(self, webhook_url: str, secret: str):
        """
        初始化 webhook 消息发送器, 暂时只使用签名方式
        :param webhook_url: webhook 的 url 地址
        :param secret: 使用签名方式，签名密钥
        """
        self._url = webhook_url
        self._secret = secret
        self._secret_enc = secret.encode('utf8')
        self._header = {
            "Content-Type": "application/json"
        }

    def _secret_params(self) -> dict:
        """
        获取签名参数
        :return: 签名参数字典
        """
        timestamp = str(round(time.time() * 1000))
        string_to_sign = f'{timestamp}\n{self._secret}'
        string_to_sign_enc = string_to_sign.encode('utf-8')
        hmac_code = hmac.new(self._secret_enc, string_to_sign_enc, digestmod=hashlib.sha256).digest()
        sign = base64.b64encode(hmac_code)
        # sign = urllib.parse.quote_plus(tmp)
        return {
            "timestamp": timestamp,
            "sign": sign
        }

    def _send(self, data: dict):
        """
        发送消息
        :param data: 要发送的消息的字典，必须满足 api 给定的格式
        """
        resp = requests.post(self._url, params=self._secret_params(), data=json.dumps(data), headers=self._header)
        print(resp.text)

    def send_text(self, txt: str):
        """
        发送文本消息
        :param txt: 文本消息的内容
        """
        self._send({
            "text": {
                "content": txt
            },
            "msgtype": "text"
        })

    def send_markdown(self, markdown: str, title: str):
        """
        发送 markdown 类型的消息
        :param title: 消息标题：消息列表看到的新消息暂时的内容
        :param markdown: markdown 格式的数据内容
        """
        self._send({
            "msgtype": "markdown",
            "markdown": {
                "title": title,
                "text": markdown
            },
        })



# from .stopwatch import StopWatch
# from .cache import cache_wrapper
# from .get_time import get_time
# from .config import Config
#
# StopWatch = StopWatch
# cache_wrapper = cache_wrapper
# get_time = get_time
# Config = Config



from typing import Optional

from tqdm import tqdm

# 当前进度/总进度|进度条|进度百分比 [用时<剩余用时, 平均用时 单位] 描述
bar_format = "{n_fmt}/{total_fmt}|{bar}|{percentage:3.0f}%［{elapsed}<{remaining},{rate_fmt}{postfix}]{desc}"


class Progress(tqdm):
    """
    使用 tqdm 包装的进度条, 修改下格式, 方便训练用.
    """

    def __init__(self, iterable=None, total=None, ncols=120):
        """
        构造函数
        Args:
            iterable: 可以对进度条进行迭代，自动推进进度条
            total: 总条目数
            ncols: 进度条总宽度, 默认120.
        """
        super().__init__(iterable=iterable, total=total, ncols=ncols, bar_format=bar_format)

    def train_result(self, loss: Optional[float] = None, acc: Optional[float] = None):
        """
        设置训练结果，同时将进度条推进 1 个条目。
        Args:
            loss: 损失函数，可空
            acc: 精确度，可空 (0-1)

        """
        desc_str = ""
        if loss is not None:
            desc_str += f" - Loss: {loss:.6f}"
        if acc is not None:
            desc_str += f" - Acc: {acc * 100:5.2f}%"
        self.set_description_str(desc_str)
        if self.iterable is None:
            # 如果不是自动迭代，推进1
            self.update(1)



import unittest
import time

from wth.utils.progress import Progress
from torch.utils.data import Dataset, DataLoader


class TestProgress(unittest.TestCase):
    def test_progress(self):
        with Progress(total=1000) as progress:
            for i in range(1000):
                time.sleep(0.01)
                progress.train_result(0.01, 0.02)

    def test_torch_progress(self):
        class _Dataset(Dataset):
            def __init__(self):
                super().__init__()

            def __len__(self):
                return 100 * 64

            def __getitem__(self, idx):
                return idx

        dataset = _Dataset()
        dataloader = DataLoader(dataset, shuffle=True, batch_size=64)

        progress = Progress(dataloader)
        for data in progress:
            time.sleep(0.02)
            progress.train_result(0.02)



from abc import ABC, abstractmethod


class BaseNotification(ABC):

    @abstractmethod
    def send_text(self, txt: str):
        pass

    @abstractmethod
    def send_markdown(self, *args, **kwargs):
        pass



import json

import requests

from .base import BaseNotification


class WechatNotification(BaseNotification):
    """
    企业微信 webhook 消息发送器.
    API 地址: https://developer.work.weixin.qq.com/document/path/91770
    """

    def __init__(self, webhook_url):
        """
        初始化器
        :param webhook_url: 机器人的 webhook 地址
        """
        self._url = webhook_url
        self._header = {
            "Content-Type": "application/json"
        }

    def send_text(self, txt: str):
        """
        发送文本消息
        :param txt: 文本消息内容
        """
        self._send({
            "msgtype": "text",
            "text": {
                "content": txt
            }
        })

    def send_markdown(self, markdown: str):
        """
        发送 markdown 格式的数据消息
        :param markdown: 消息内容
        """
        self._send({
            "msgtype": "markdown",
            "markdown": {
                "content": markdown
            }
        })

    def _send(self, data: dict):
        """
        发送消息
        :param data: 企业微信的数据 body
        """
        resp = requests.post(self._url, data=json.dumps(data), headers=self._header)
        print(resp.text)



import torch

from .utils import try_gpu, set_device_local, set_dtype_local
from ..utils.config import Config as _Config


class Config(_Config):
    """
    环境配置，使用 yml 文件。
    例如:
    ```
    active: local
      school-server:
      local:
        epochs: 10
        batch-size: 64
        dtype: float32
        cuda_no: 1
        dataset_path: /your/dataset/path
    ```
    active 必需标识启用的环境。
    后续使用 key: 环境配置的方式可以定义多个环境。
    """

    def __init__(self, config_file='config.yml'):
        """
        构造函数。
        Args:
            config_file: 配置文件所在位置，默认 config.yml。
        """
        super(Config, self).__init__(config_file=config_file)
        self.epochs = self.configs.get('epochs', 10)
        self.batch_size = self.configs.get('batch-size', 64)
        self.dtype = self.configs.get('dtype', 'float32')
        self.cuda_no = self.configs.get('cuda-no')
        self.dataset_path = self.configs.get('dataset-path')

        # 设置设备
        if self.cuda_no is not None:
            set_device_local(try_gpu(self.cuda_no))
        # 设置 dtype
        if self.dtype is not None:
            if self.dtype == 'float32':
                set_dtype_local(torch.float32)
            else:
                set_dtype_local(torch.float64)



from wth.utils.config import Config

try:
    # 可选模块 ~~~
    import pytorch_lightning as pl


    def _pl_config(config: Config) -> dict:
        return config['pl']


    def trainer(config: Config) -> pl.Trainer:
        pl_config = _pl_config(config)
        if pl_config is None:
            raise KeyError(f"请在{config.config_file}中定义 pl")
        devices = pl_config['devices']

        return pl.Trainer()

except ImportError:
    print("Please install pytorch-lightning")



from wth.utils import cache_wrapper


@cache_wrapper(prefix="test_utils_cache")
def cache_1():
    return 1


@cache_wrapper(prefix="test_utils_cache")
def cache_2(a):
    return a


@cache_wrapper(prefix="test_utils_cache")
def cache_3(a=1):
    return {'a': a}


@cache_wrapper(prefix="test_utils_cache")
def cache_4(a, b=2):
    return {'a': a, 'b': b}


cache_1()
cache_2(2)
cache_3(a=3)
cache_4(4, b=4)



import json
import sys

import yaml


class Config:
    """
    环境配置，使用 yml 文件。
    例如:
    ```
    active: local
    school-server:
    local:
    shared:
    ```
    active 必需标识启用的环境。
    后续使用 key: 环境配置的方式可以定义多个环境。
    shared: 是共用的配置。
    """

    def __init__(self, config_file='config.yml'):
        """
        构造函数。
        Args:
            config_file: 配置文件所在位置，默认 config.yml。
        """
        self.config_file = config_file
        with open(config_file) as f:
            self.original = yaml.load(f, Loader=yaml.FullLoader)
            self.active_profile = self.original['active']
            if self.active_profile is None:
                raise ValueError("No active profile")

            shared_key = 'shared'

            if self.active_profile == shared_key:
                raise ValueError("[shared] is a shared configuration.")

            print(f"Active profile[{self.active_profile}].")
            sys.stdout.flush()

            self.configs = self.original[self.active_profile]  # type: dict
            if self.configs is None:
                print(f"Active profile[{self.active_profile}] has no arguments.", file=sys.stderr)
                sys.stderr.flush()
                self.configs = {}

            # 添加共用部分, 如果存在一样的 key, 则使用 active 激活的配置内的
            if shared_key in self.original:
                for k, v in self.original[shared_key].items():
                    # 如果 激活的配置中没有 shared 内的定义, 则添加
                    if k not in self.configs:
                        self.configs[k] = v
                    else:
                        print(f'key[{k}] both in (shared, {self.active_profile}), use ({self.active_profile})')

    def __getitem__(self, item):
        if self.configs.__contains__(item):
            return self.configs[item]
        key = item.replace("_", "-")
        if self.configs.__contains__(key):
            return self.configs[key]
        return None

    def __str__(self):
        return json.dumps(self.original)
